# main.py

from __future__ import division

import getopt

import pygame
from OpenGL.GL import *
from OpenGL.GLU import *

import scenery, heroism, texturing, controls


VERSION = "Pyweek 1.01"

LONG_OPTIONS = ('help','version','fullscreen','level=','framerate=',
                'up=','down=','left=','right=','jump=','attack=')
OPTIONS = "hvfl:r:x:"

USAGE = """\
usage: steam [options]
options:
  -h,--help            Print this help screen
  -v,--version         Print version information
  -f,--fullscreen      Full screen mode
  -l #,--level=#       Set level number
  -r #,--framerate=#   Set number of frames per second
  --up=key             Set key/js axis for up
  --down=key           Set key/js axis for down
  --left=key           Set key/js axis for left
  --right=key          Set key/js axis for right
  --jump=key           Set key/js button for jump
  --attack=key         Set key/js button for attack
  --quit=key           Set key/js button for quit

"key" can be a Pygame (aka SDL) keycode without the K_ prefix,
or it can be of the form js# to map a certain joystick button
or axis.
"""


def main(*args):
    fullscreen = False
    frames_per_second = 20
    level = '1'
    lvlfile = None

    try:
        opts, rest = getopt.getopt(args, OPTIONS, LONG_OPTIONS)
    except getopt.GetoptError:
        print "error: unknown option"
        print USAGE
        raise SystemExit(1)

    for opt,arg in opts:
        if opt in ('-h', '--help'):
            print USAGE
            raise SystemExit
        if opt in ('-v', '--version'):
            print "Version: %s" % VERSION
            raise SystemExit
        if opt in ('-f', '--fullscreen'):
            fullscreen = True
        elif opt in ('-l','--level'):
            level = arg
        elif opt in ('-r','--framerate'):
            frames_per_second = int(arg)
        elif opt in ('--left','--right','--up','--down'):
            controls.mapkey(opt[2:],arg,True)
        elif opt in ('--jump','--attack','--quit'):
            controls.mapkey(opt[2:],arg,False)
        elif opt == '-x':
            import os
            lvlfile = os.path.basename(arg)

    if lvlfile is None:
        lvlfile = "level%s.xml" % level

    modeflags = pygame.OPENGL|pygame.DOUBLEBUF
    if fullscreen:
        modeflags |= pygame.FULLSCREEN

    pygame.init()
    pygame.display.set_mode((640,480),modeflags)
    pygame.display.set_caption('Blowing off Steam')
    pygame.mouse.set_visible(False)
    pygame.mouse.set_pos(1,1)

    clock = pygame.time.Clock()

    if pygame.joystick.get_count() > 0:
        pygame.joystick.Joystick(0).init()

    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA)

    heroine = heroism.Heroine()

    while True:
        run_title_screen(heroine,clock,frames_per_second)
        run_blank(1.0,clock,frames_per_second)
        run_scene(lvlfile,heroine,clock,frames_per_second)
        run_blank(4.0,clock,frames_per_second,heroine.kills)


def run_title_screen(heroine,clock,frames_per_second):
    scene = scenery.Scene('title.xml')
    scene.place_heroine(heroine)
    heroine.inair = False
    heroine.steamtimer = 0.0
    scene.view_x = 30.0
    scene.view_y = 15.0
    gamestart = False
    pygame.event.get()
    while True:
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                f = controls.KEYMAP.get(event.key)
                if f in ("jump","attack"):
                    gamestart = True
                elif f == "quit":
                    raise SystemExit
            elif event.type == pygame.JOYBUTTONDOWN:
                f = controls.JOYBUTTONMAP.get(event.button)
                if f in ("jump","attack"):
                    gamestart = True
                elif f == "quit":
                    raise SystemExit
            elif event.type == pygame.QUIT:
                raise SystemExit
        if gamestart:
            break
        scene.draw()
        clock.tick(frames_per_second)
        pygame.display.flip()
    scene.unload()


def run_blank(blanktime,clock,frames_per_second,kills=None):
    if kills is not None:
        tk = texturing.Texture('totalkills.png')
        nb = texturing.Texture('whitenumbers.png')
    dt = 1.0/frames_per_second
    pygame.event.get()
    while True:
        blanktime -= dt
        if blanktime <= 0:
            break
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                f = controls.KEYMAP.get(event.key)
                if f == "quit":
                    raise SystemExit
            elif event.type == pygame.JOYBUTTONDOWN:
                f = controls.JOYBUTTONMAP.get(event.button)
                if f == "quit":
                    raise SystemExit
            elif event.type == pygame.QUIT:
                raise SystemExit
        glClearColor(0.0,0.0,0.0,1.0)
        glClear(GL_COLOR_BUFFER_BIT)
        if kills is not None:
            glLoadIdentity()
            gluOrtho2D(0.0,40.0,0.0,30.0)
            glColor(1.0,1.0,1.0,1.0)
            nb.bind()
            glBegin(GL_QUADS)
            s = str(kills)
            x = 20.0 - 1.6*len(s)
            for c in s:
                n = ord(c)-48
                if not 0 <= n <= 9:
                    continue
                glTexCoord(0.1*n,0.0)
                glVertex(x,11.0)
                glTexCoord(0.1*(n+1),0.0)
                glVertex(x+3.2,11.0)
                glTexCoord(0.1*(n+1),1.0)
                glVertex(x+3.2,15.0)
                glTexCoord(0.1*n,1.0)
                glVertex(x,15.0)
                x += 3.2
            glEnd()        
            nb.unbind()
            tk.bind()
            glBegin(GL_QUADS)
            glTexCoord(0.0,0.0)
            glVertex(15.0,17.0)
            glTexCoord(1.0,0.0)
            glVertex(25.0,17.0)
            glTexCoord(1.0,1.0)
            glVertex(25.0,21.0)
            glTexCoord(0.0,1.0)
            glVertex(15.0,21.0)
            glEnd()
            tk.unbind()
        clock.tick(frames_per_second)
        pygame.display.flip()
    if kills is not None:
        tk.unload()
        nb.unload()


def run_scene(level,heroine,clock,frames_per_second):
    scene = scenery.Scene(level)
    scene.place_heroine(heroine)
    dt = 1.0/frames_per_second
    pygame.event.get()
    while True:
        heroine.process_input()
        status = scene.poll_interactions()
        if not status:
            break
        scene.update(dt)
        scene.draw()
        heroine.draw_overlay()
        clock.tick(frames_per_second)
        pygame.display.flip()
    scene.unload()
