# scenery.py

from __future__ import division

import os
import math
import random

from elementtree import ElementTree as ET
from OpenGL.GL import *
from OpenGL.GLU import *

import texturing, enmity

def _import_numpy():
    try:
        import numpy as N
        return N
    except ImportError:
        pass
    try:
        import Numeric as N
        return N
    except ImportError:
        pass
    try:
        import numarray as N
        return N
    except ImportError:
        pass
    raise RuntimeError("could not import Numeric, numarray, nor numpy")

N = _import_numpy()


class Scene(object):
    HORIZ_MARGIN = 17.0
    VERT_MARGIN = 6.0
    VIEW_WIDTH = 40.0
    VIEW_HEIGHT = 30.0
    GROUNDEPS = 0.00001
    GROUNDPROT = GROUNDEPS
    JUMPEPS = 0.001
    MAX_ENEMIES = 7

    def __init__(self,filename):
        xscene = ET.parse(os.path.join('level',filename)).getroot()

        # Read in some scene parameters

        self.width = float(xscene.find('width').text)
        self.height = float(xscene.find('height').text)

        # Create various fixed-data containers

        self.textures = {}

        self.ground_x1 = []
        self.ground_y1 = []
        self.ground_x2 = []
        self.ground_y2 = []
        self.ground_flags = []

        self.background = None

        self.apples = []
        self.apple_image = None
        self.espressos = []
        self.espresso_image = None

        banners = []
        fences = []
        trees = []
        landscapes = []
        objects = []

        # Read in list of coordinates

        x_points = [ float(x) for x in xscene.find('points/x').text.split(',') ]
        y_points = [ float(x) for x in xscene.find('points/y').text.split(',') ]

        # Read in entrance point

        si = int(xscene.find('entrance').text)
        self.entrance = x_points[si],y_points[si]

        # Fill in ground coordinate and flag arrays
        # Load any ground textures encountered

        for xground in xscene.findall('ground'):
            seq = [ int(n) for n in xground.find('seq').text.split(',') ]
            icy = xground.find('icy').text == 'yes'
            jd = xground.find('jumpdown').text == 'yes'
            for si1, si2 in zip(seq[:-1],seq[1:]):
                self.ground_x1.append(x_points[si1])
                self.ground_y1.append(y_points[si1])
                self.ground_x2.append(x_points[si2])
                self.ground_y2.append(y_points[si2])
                self.ground_flags.append(1 | (jd << 1) | (icy << 2))
            self.ground_flags[-1] &= ~1
            xbanner = xground.find('banner')
            if xbanner is not None:
                texname = xbanner.text
                if texname not in self.textures:
                    self.textures[texname] = texturing.Texture(texname)
                pointlist = []
                for si in seq:
                    pointlist.append((x_points[si],y_points[si]))
                banners.append((pointlist,texname))

        # Change lists into arrays

        self.ground_x1 = N.array(self.ground_x1,N.Float)
        self.ground_y1 = N.array(self.ground_y1,N.Float)
        self.ground_x2 = N.array(self.ground_x2,N.Float)
        self.ground_y2 = N.array(self.ground_y2,N.Float)
        self.ground_flags = N.array(self.ground_flags,N.UnsignedInteger)

        # Read in a list of landscapes (for the lack of a better name)

        for xlandscape in xscene.findall('landscape'):
            seq = [ int(n) for n in xlandscape.find('seq').text.split(',') ]
            pointlist = []
            for si in seq:
                pointlist.append((x_points[si],y_points[si]))
            texname = xlandscape.find('background').text
            if texname not in self.textures:
                self.textures[texname] = texturing.Texture(texname)
            scale = float(xlandscape.find('scale').text)
            landscapes.append((pointlist,texname,scale))

        # Read in a list of fences

        for xfence in xscene.findall('fence'):
            seq = [ int(n) for n in xfence.find('seq').text.split(',') ]
            pointlist = []
            for si in seq:
                pointlist.append((x_points[si],y_points[si]))
            texname = xfence.find('banner').text
            if texname not in self.textures:
                self.textures[texname] = texturing.Texture(texname)
            fences.append((pointlist,texname))

        # Read in a list of trees

        for xtrees in xscene.findall('trees'):
            seq = [ int(n) for n in xtrees.find('seq').text.split(',') ]
            pointlist = []
            for si in seq:
                pointlist.append((x_points[si],y_points[si]))
            texname = xtrees.find('banner').text
            if texname not in self.textures:
                self.textures[texname] = texturing.Texture(texname)
            trees.append((pointlist,texname))

        # Read in a list of apples

        for xapples in xscene.findall('apples'):
            seq = [ int(n) for n in xapples.find('seq').text.split(',') ]
            for si in seq:
                self.apples.append((x_points[si],y_points[si]))
            texname = xapples.find('image').text
            if texname not in self.textures:
                self.textures[texname] = texturing.Texture(texname)
            self.apple_image = texname

        # Read in a list of espressos

        for xespressos in xscene.findall('espressos'):
            seq = [ int(n) for n in xespressos.find('seq').text.split(',') ]
            for si in seq:
                self.espressos.append((x_points[si],y_points[si]))
            texname = xespressos.find('image').text
            if texname not in self.textures:
                self.textures[texname] = texturing.Texture(texname)
            self.espresso_image = texname

        # Read in a list of objects

        for xobject in xscene.findall('object'):
            bl = int(xobject.find('bl').text)
            tr = int(xobject.find('tr').text)
            l = x_points[bl]
            b = y_points[bl]
            r = x_points[tr]
            t = y_points[tr]
            pointlist = ((l,b),(r,b),(r,t),(l,t))
            texname = xobject.find('image').text
            if texname not in self.textures:
                self.textures[texname] = texturing.Texture(texname)
            objects.append((pointlist,texname))

        # Read in background image

        xbackground = xscene.find('background')
        if xbackground is not None:
            texname = xbackground.text
            if texname not in self.textures:
                self.textures[texname] = texturing.Texture(texname)
            self.background = texname

        # Create display list for the static scene

        self.static_dl = glGenLists(1)
        if not self.static_dl:
           raise RuntimeError("can't allocate display list")
        glNewList(self.static_dl,GL_COMPILE)
        for pointlist,texname,scale in landscapes:
            self._draw_landscape(pointlist,texname,scale)
        for pointlist,texname in trees:
            self._draw_trees(pointlist,texname)
        for pointlist,texname in fences:
            self._draw_fence(pointlist,texname)
        for pointlist,texname in objects:
            self._draw_object(pointlist,texname)
        for pointlist,texname in banners:
            self._draw_banner(pointlist,texname)
        glEndList()

        # Scene state

        self.event_timer = 0.0
        self.enemy_time = 0.0
        self.view_x = 0.0
        self.view_y = 0.0

        # Dynamic objects (characters and items)

        self.heroine = None
        self.enemies = []
 
    def unload(self):
        glDeleteLists(self.static_dl,1)
        if self.heroine is not None:
            self.heroine.exit_scene()
        for enemy in self.enemies:
            enemy.exit_scene()
            enemy.unload()
        for texture in self.textures.itervalues():
            texture.unload()
        self.__dict__.clear()

    def place_heroine(self,heroine):
        self.heroine = heroine
        self.heroine.enter_scene(self,self.entrance[0],self.entrance[1])

    def create_enemy(self):
        if len(self.enemies) >= self.MAX_ENEMIES:
            return
        if (self.heroine.facing == "right") == (random.random() < 0.8):
            xe = self.heroine.x + 40.0
        else:
            xe = self.heroine.x - 40.0
        gs = N.nonzero(N.logical_and(
            self.ground_x1 <= xe, xe <= self.ground_x2))
        if len(gs) == 0:
            return
        elif len(gs) == 1:
            g = gs[0]
        else:
            g = random.choice(g)
        x1 = self.ground_x1[g]
        y1 = self.ground_y1[g]
        x2 = self.ground_x2[g]
        y2 = self.ground_y2[g]
        ye = (y1*(x2-xe)+y2*(xe-x1))/(x2-x1)
        ntp = random.choice((1,2,3))
        if ntp == 1:
            enemy = enmity.FreeNinja()
            enemy.enter_scene(self,xe,ye)
        elif ntp == 2:
            vary = random.uniform(10.0,30.0)
            enemy = enmity.BoundNinja()
            enemy.enter_scene(self,xe,ye,xe-vary,xe+vary)
        else:
            enemy = enmity.JumpingNinja()
            enemy.enter_scene(self,xe,ye)
        self.enemies.append(enemy)

    def grab_object(self,x,y):
        rvapples = 0
        napples = []
        for xa,ya in self.apples:
            dx = x-xa
            dy = y-ya
            if -2.0 <= dx <= 2.0 and -5.5 <= dy <= 2.0:
                rvapples += 1
            else:
                napples.append((xa,ya))
        self.apples = napples
        rvespressos = 0
        nespressos = []
        for xa,ya in self.espressos:
            dx = x-xa
            dy = y-ya
            if -2.0 <= dx <= 2.0 and -5.5 <= dy <= 2.0:
                rvespressos += 1
            else:
                nespressos.append((xa,ya))
        self.espressos = nespressos
        return rvapples, rvespressos

    def poll_interactions(self):
        if self.event_timer > self.enemy_time:
            self.create_enemy()
            self.enemy_time = self.event_timer + random.normalvariate(2.0,1.0)
        for enemy in self.enemies:
            enemy.poll_interactions()
        self.heroine.poll_interactions()
        if self.heroine.life <= 0 or self.heroine.y <= 0:
            self.heroine.exit_scene()
            self.heroine = None
            return False
        nenemies = []
        for enemy in self.enemies:
            if (enemy.life <= 0 or enemy.y <= 0
                or abs(enemy.y-self.heroine.y) > 50.0):
                enemy.exit_scene()
                enemy.unload()
            else:
                nenemies.append(enemy)
        self.enemies = nenemies
        return True

    def update(self,dt):
        for enemy in self.enemies:
            enemy.update(dt)
        self.heroine.update(dt)
        leftedge = max(0.0,self.heroine.x-self.HORIZ_MARGIN)
        rightedge = min(self.width-self.VIEW_WIDTH,
                        self.heroine.x+self.HORIZ_MARGIN-self.VIEW_WIDTH)
        bottomedge = max(0.0,self.heroine.y-self.VERT_MARGIN)
        topedge = min(self.height-self.VIEW_HEIGHT,
                      (self.heroine.y+self.heroine.HEIGHT
                       +self.VERT_MARGIN-self.VIEW_HEIGHT))
        self.view_x = min(leftedge,max(rightedge,self.view_x))
        self.view_y = min(bottomedge,max(topedge,self.view_y))
        self.event_timer += dt

    def draw(self):
        glClearColor(0.8,0.8,1.0,0.0)
        glClear(GL_COLOR_BUFFER_BIT)
        glLoadIdentity()
        gluOrtho2D(self.view_x,self.view_x+self.VIEW_WIDTH,
                   self.view_y,self.view_y+self.VIEW_HEIGHT)
        if self.background is not None:
            self.draw_background()
        glCallList(self.static_dl)
        for x,y in self.apples:
            self.draw_icon(x,y,self.apple_image)
        for x,y in self.espressos:
            self.draw_icon(x,y,self.espresso_image)
        for enemy in self.enemies:
            enemy.draw()
        if self.heroine is not None:
            self.heroine.draw()

    def draw_background(self):
        self.textures[self.background].bind()
        xtc0 = self.view_x/250.0
        glBegin(GL_QUADS)
        glColor(1.0,1.0,1.0,1.0)
        glTexCoord(xtc0,0.0)
        glVertex(self.view_x,0.9*self.view_y)
        glTexCoord(xtc0+0.7,0.0)
        glVertex(self.view_x+self.VIEW_WIDTH,0.9*self.view_y)
        glTexCoord(xtc0+0.7,0.98)
        glVertex(self.view_x+self.VIEW_WIDTH,
                 0.9*self.view_y+self.VIEW_HEIGHT*0.65)
        glTexCoord(xtc0,0.98)
        glVertex(self.view_x,0.9*self.view_y+self.VIEW_HEIGHT*0.65)
        glEnd()
        self.textures[self.background].unbind()        

    def draw_icon(self,x,y,texname):
        self.textures[texname].bind()
        glBegin(GL_QUADS)
        glColor(1.0,1.0,1.0,1.0)
        glTexCoord(0.0,0.0)
        glVertex(x-1.0,y)
        glTexCoord(1.0,0.0)
        glVertex(x+1.0,y)
        glTexCoord(1.0,1.0)
        glVertex(x+1.0,y+2.0)
        glTexCoord(0.0,1.0)
        glVertex(x-1.0,y+2.0)
        glEnd()
        self.textures[texname].unbind()

    def _draw_banner(self,pointlist,texname):
        self.textures[texname].bind()
        glBegin(GL_QUAD_STRIP)
        glColor(1.0,1.0,1.0,1.0)
        for x,y in pointlist:
            glTexCoord(x/16.0,1.0)
            glVertex(x,y+0.2)
            glTexCoord(x/16.0,0.0)
            glVertex(x,y-1.8)
        glEnd()
        self.textures[texname].unbind()

    def _draw_fence(self,pointlist,texname):
        self.textures[texname].bind()
        x0 = pointlist[0][0]
        glBegin(GL_QUAD_STRIP)
        glColor(1.0,1.0,1.0,1.0)
        for x,y in pointlist[:-1]:
            glTexCoord(0.5*(x-x0),1.0)
            glVertex(x,y+4.0)
            glTexCoord(0.5*(x-x0),0.0)
            glVertex(x,y)
        x,y =  pointlist[-1]
        xl = int(0.5*(x-x0)) + 0.85
        glTexCoord(xl,1.0)
        glVertex(x,y+4.0)
        glTexCoord(xl,0.0)
        glVertex(x,y)
        glEnd()
        self.textures[texname].unbind()

    def _draw_trees(self,pointlist,texname):
        self.textures[texname].bind()
        glBegin(GL_QUADS)
        glColor(1.0,1.0,1.0,1.0)
        for x,y in pointlist:
            glTexCoord(0.0,0.0)
            glVertex(x-8.0,y-1.0)
            glTexCoord(1.0,0.0)
            glVertex(x+8.0,y-1.0)
            glTexCoord(1.0,1.0)
            glVertex(x+8.0,y+20.0)
            glTexCoord(0.0,1.0)
            glVertex(x-8.0,y+20.0)
        glEnd()
        self.textures[texname].unbind()

    def _draw_object(self,pointlist,texname):
        self.textures[texname].bind()
        glBegin(GL_QUADS)
        glColor(1.0,1.0,1.0,1.0)
        glTexCoord(0.0,0.0)
        glVertex(pointlist[0][0],pointlist[0][1])
        glTexCoord(1.0,0.0)
        glVertex(pointlist[1][0],pointlist[1][1])
        glTexCoord(1.0,1.0)
        glVertex(pointlist[2][0],pointlist[2][1])
        glTexCoord(0.0,1.0)
        glVertex(pointlist[3][0],pointlist[3][1])
        glEnd()
        self.textures[texname].unbind()

    def _draw_landscape(self,pointlist,texname,scale):
        self.textures[texname].bind()
        glBegin(GL_QUAD_STRIP)
        glColor(1.0,1.0,1.0,1.0)
        for x,y in pointlist:
            glTexCoord(x*scale,y*scale)
            glVertex(x,y)
            glTexCoord(x*scale,0.0)
            glVertex(x,0.0)
        glEnd()
        self.textures[texname].unbind()
        glBegin(GL_LINE_STRIP)
        glColor(0.0,0.0,0.0)
        glVertex(pointlist[0][0],0.0)
        for x,y in pointlist:
            glVertex(x,y)
        glVertex(pointlist[-1][0],0.0)
        glEnd()

    def move(self,x0,y0,dx,dy,jumpdown=False,checkground=True):
        xf = x0+dx
        yf = y0+dy
        x1 = self.ground_x1
        y1 = self.ground_y1
        x2 = self.ground_x2
        y2 = self.ground_y2
        y0g = (y2*(x0-x1)+y1*(x2-x0))/(x2-x1)

        if checkground:
            onground = N.nonzero(N.logical_and.reduce((
                x0 >= x1, x0 <= x2, abs(y0-y0g) < self.GROUNDEPS)))
            if len(onground) > 0:
                g = onground[0]
                if (abs(dy) < self.JUMPEPS):
                    return self.run(g,x0,dx)
                if not self.ground_flags[g] & 2:
                    yfg = (y2[g]*(xf-x1[g])+y1[g]*(x2[g]-xf))/(x2[g]-x1[g])
                    if yf < yfg:
                        return self.run(g,x0,dx)

        yfg = (y2*(xf-x1)+y1*(x2-xf))/(x2-x1)

        crossmask = N.logical_and(
            y0-y0g >= self.GROUNDEPS, yf-yfg < self.GROUNDEPS)
        if jumpdown:
            bitmask = (self.ground_flags & 2).astype(N.UnsignedInteger)
            crossmask = N.logical_and(crossmask,N.logical_not(bitmask))
        crossing = N.nonzero(crossmask)
        if len(crossing) > 0:
            y0gT = N.take(y0g,crossing)
            yfgT = N.take(yfg,crossing)
            x1T = N.take(x1,crossing)
            x2T = N.take(x2,crossing)                
            xc = (xf*(y0-y0gT) - x0*(yf-yfgT)) / (y0-y0gT-yf+yfgT)
            toground = N.nonzero(N.logical_and(
                xc > x1T-self.GROUNDPROT, xc < x2T+self.GROUNDPROT))
            if len(toground) > 0:
                g = toground[0]
                xc = xc[g]
                y0gT = y0gT[g]
                yfgT = yfgT[g]
                yc = (yf*(y0-y0gT) - y0*(yf-yfgT)) / (y0-y0gT-yf+yfgT)
                return xc,yc,False

        return x0+dx,y0+dy,True
        
    def run(self,g,x0,dx):
        xf = x0+dx
        while xf > self.ground_x2[g]:
            if self.ground_flags[g] & 1:
                g += 1
            else:
                tl = 1.0 - (self.ground_x2[g] - x0) / dx
                return self.move(self.ground_x2[g],self.ground_y2[g],
                                 dx*tl,0.0,False,False)
        while xf < self.ground_x1[g]:
            if g > 0 and self.ground_flags[g-1] & 1:
                g -= 1
            else:
                tl = 1.0 - (xf - self.ground_x1[g]) / dx
                return self.move(self.ground_x1[g],self.ground_y1[g],
                                 dx*tl,0.0,False,False)
        dyg = self.ground_y2[g]-self.ground_y1[g]
        dxg = self.ground_x2[g]-self.ground_x1[g]
        yf = self.ground_y1[g] + (xf-self.ground_x1[g])*dyg/dxg
        return xf,yf,False

    def attack(self,r0,r1,a0,a1,power,attack):
        hits = 0
        kills = 0
        r0sq = r0*r0
        r1sq = r1*r1
        for enemy in self.enemies:
            dx = enemy.x-self.heroine.x
            dy = enemy.y-self.heroine.y
            dsq = dx*dx+dy*dy
            if dsq < r0sq or dsq > r1sq:
                continue
            th = math.atan2(dy,dx)
            if a1 >= a0:
                if not (a1 >= th >= a0):
                    continue
            else:
                if a0 > th > a1:
                    continue
            hits += 1
            enemy.life -= power
            if enemy.life <= 0:
                kills += 1
            enemy.recoil(self.heroine.x,self.heroine.y)
        return hits,kills

    def be_attacked(self,xr,yr):
        for enemy in self.enemies:
            xdiff = abs(enemy.x-self.heroine.x)
            ydiff = abs(enemy.y-self.heroine.y)
            if xdiff <= xr and ydiff <= yr:
                self.heroine.hurt(enemy)
                self.heroine.recoil(enemy.x,enemy.y)
        
                
