# terrain.py

from __future__ import division

import itertools
import logging
import random
import math
import sys
import os

from elementtree import ElementTree as ET


logger = logging.getLogger('terrain')
logger.addHandler(logging.StreamHandler(sys.stderr))
#logger.setLevel(logging.ERROR)
logger.setLevel(logging.DEBUG)


VERSION = "0007"

LENGTH_MAX = 1000.0

N_SUBLEVELS_MIN = 1
N_SUBLEVELS_MAX = 5
FLAT_SUBLEVEL_PCT = 0.8
MAX_SUBLEVEL_ANGLE = 0.5

MIN_PLATFORM = 6.0

N_GAP_CHOICE = [0,0,0,0,1,1,2]
MIN_GAP_WIDTH = 5.0
MAX_GAP_WIDTH = 10.0

MIN_AVG_SEG_IVAL = 10.0
MAX_AVG_SEG_IVAL = 50.0
MIN_ABS_SEG_IVAL = 3.0
SIGMA_SEG_IVAL = 5.0

SIGMA_SIGMA_HILL = 4.0
SIGMA_ENDPOINT = 1.0

NOGAP_PCT = 0.5
NOHILS_PCT = 0.05

ENDPOINT_MIN_SEP = 6.0
ENDPOINT_MIN_SEP_SQ = ENDPOINT_MIN_SEP**2

TREES_PER_FOOT = 0.04
TREE_SLOPE_MAX = 0.25
FENCE_PCT = 0.20
FENCE_SEG_MAX = 50.0
FENCE_SLOPE_MAX = 0.25
GOODIES_PER_FOOT = 0.01
GOODIE_SLOPE_MAX = 0.25
APPLE_TREE_RATIO = 0.1

BOUND_NINJAS_PER_FOOT = 0.04
FREE_NINJAS_PER_FOOT = 0.04



class TerrainGenerator(object):

    def __init__(self,seed):
        seed = int(seed)
        
        self.xmlroot = ET.Element("scene")
        self.xmlroot.attrib["id"] = "level%d-%s" % (seed,VERSION)

        self.points = []
        self.pointhash = {}
        self.endpoints = {}
        self.trees = []
        self.espressos = []
        self.apples = []
        self.free_ninjas = []
        self.bound_ninjas = []

        random.seed(int(seed))

        nsublevels = random.choice(range(N_SUBLEVELS_MIN,N_SUBLEVELS_MAX+1))

        slvllength_max = LENGTH_MAX / nsublevels
        slvllength_min = slvllength_max / 4
        slvllengths = [ random.uniform(slvllength_min,slvllength_max)
                        for i in xrange(nsublevels) ]

        slvlangles = []
        for i in xrange(nsublevels):
            if random.random() < FLAT_SUBLEVEL_PCT:
                slvlangles.append(0)
            else:
                slvlangles.append(
                    random.uniform(-MAX_SUBLEVEL_ANGLE,MAX_SUBLEVEL_ANGLE))

        slvlpoints = [(5.0,0.0)]
        miny = 0.0
        maxy = 0.0
        for length,angle in zip(slvllengths,slvlangles):
            x0,y0 = slvlpoints[-1]
            x1 = x0 + length*math.cos(angle)
            y1 = y0 + length*math.sin(angle)
            slvlpoints.append((x1,y1))
            if y1 < miny:
                miny = y1
            elif y1 < maxy:
                maxy = y1

        for (x0,y0),(x1,y1),i in zip(slvlpoints[:-1],slvlpoints[1:],
                                     itertools.count()):
            self.baseland(x0,y0,x1-MIN_GAP_WIDTH,y1,i==0)

        if self.trees:
            xtrees = ET.SubElement(self.xmlroot,"trees")
            xseq = ET.SubElement(xtrees,"seq")
            xseq.text = ",".join(str(x) for x in self.trees)
            xbanner = ET.SubElement(xtrees,"banner")
            xbanner.text = "tree.png"

        if self.apples:
            xapples = ET.SubElement(self.xmlroot,"apples")
            xseq = ET.SubElement(xapples,"seq")
            xseq.text = ",".join(str(x) for x in self.apples)
            ximage = ET.SubElement(xapples,"image")
            ximage.text = "apple.png"

        if self.espressos:
            xespressos = ET.SubElement(self.xmlroot,"espressos")
            xseq = ET.SubElement(xespressos,"seq")
            xseq.text = ",".join(str(x) for x in self.espressos)
            ximage = ET.SubElement(xespressos,"image")
            ximage.text = "espresso.png"

        #xfreeninjas = ET.SubElement(self.xmlroot,"freeninjas")
        #xseq = ET.SubElement(xfreeninjas,"seq")
        #xseq.text = ",".join(str(x) for x in self.free_ninjas)

        #xboundninjas = ET.SubElement(self.xmlroot,"boundninjas")
        #xseq = ET.SubElement(xboundninjas,"seq")
        #xseq.text = ",".join(str(x[0]) for x in self.bound_ninjas)
        #xminr = ET.SubElement(xboundninjas,"minrange")
        #xminr.text = ",".join(str(x[1]) for x in self.bound_ninjas)
        #xmaxr = ET.SubElement(xboundninjas,"maxrange")
        #xmaxr.text = ",".join(str(x[2]) for x in self.bound_ninjas)

        minx = self.points[0][0]
        maxx = self.points[0][0]
        miny = self.points[0][1]
        maxy = self.points[0][1]
        for x,y in self.points:
            if x < minx:
                minx = x
            elif x > maxx:
                maxx = x
            if y < miny:
                miny = y
            elif y > maxy:
                maxy = y

        yoff = -miny+5.0
        xoff = -minx+5.0
        width = maxx-minx+10.0
        height = maxy-miny+50.0

        xpoints = ET.SubElement(self.xmlroot,"points")
        xps =  ",".join(str(x+xoff) for x,y in self.points)
        yps = ",".join(str(y+yoff) for x,y in self.points)
        ET.SubElement(xpoints,"x").text = xps
        ET.SubElement(xpoints,"y").text = yps
                
        ET.SubElement(self.xmlroot,"width").text = str(width)
        ET.SubElement(self.xmlroot,"height").text = str(height)

        ET.SubElement(self.xmlroot,"background").text='mountain-background.png'
        
        filename = os.path.join("xml","%s.xml"%self.xmlroot.attrib["id"])
        ET.ElementTree(self.xmlroot).write(filename)


    def pointindex(self,x,y):
        index = self.pointhash.get((x,y))
        if index is None:
            index = len(self.points)
            self.pointhash[(x,y)] = index
            self.points.append((x,y))
        return index


    def adjust_endpoint(self,x,y):
        xbin = int(x//10.0)
        ybin = int(y//10.0)
        ptc = [ (xe,ye)
                for i in range(xbin-1,xbin+2)
                for j in range(ybin-1,ybin+2)
                for xe,ye in self.endpoints.get((i,j),()) ]
        for xe,ye in ptc:
            dx = x-xe
            dy = y-ye
            dsq = dx*dx + dy*dy
            if dsq < ENDPOINT_MIN_SEP_SQ:
                d = math.sqrt(dsq)
                if d == 0:
                    y = y-ENDPOINT_MIN_SEP
                    ybin = int(y//10.0)
                else:
                    x = xe + dx*ENDPOINT_MIN_SEP/d
                    y = ye + dy*ENDPOINT_MIN_SEP/d
                    xbin = int(x//10.0)
                    ybin = int(y//10.0)
                break
        if (xbin,ybin) in self.endpoints:
            self.endpoints[xbin,ybin].append((x,y))
        else:
            self.endpoints[xbin,ybin] = [(x,y)]
        return x,y


    def intervalize(self,x0,x1,avgseg,sigma,absminseg):
        if x1-x0 > absminseg and random.uniform(avgseg,sigma) < x1-x0:
            xc = random.uniform((x0+x0+x1)/3,(x0+x1+x1)/3)
            before = self.intervalize(x0,xc,avgseg,sigma,absminseg)
            after = self.intervalize(xc,x1,avgseg,sigma,absminseg)
            return before+[xc]+after
        else:
            return []


    def decorate_segment(self,x0,y0,x1,y1):
        seglen = x1-x0
        segheight = abs(y0-y1)
        if (seglen <= FENCE_SEG_MAX and segheight/seglen < FENCE_SLOPE_MAX
            and random.random() < FENCE_PCT):
            logger.debug("            Fence: (%s,%s)-(%s,%s)" % (x0,y0,x1,y1))
            xfence = ET.SubElement(self.xmlroot,"fence")
            xseq = ET.SubElement(xfence,"seq")
            si0 = self.pointindex(x0,y0)
            si1 = self.pointindex(x1,y1)
            xseq.text = "%d,%d" % (si0,si1)
            xbanner = ET.SubElement(xfence,"banner")
            xbanner.text = "fence.png"
        f = 0.5 * math.pi
        if segheight/seglen < TREE_SLOPE_MAX:
            thresh = min(math.atan(TREES_PER_FOOT*seglen*f)/f,0.9)
            while random.random() < thresh:
                xt = random.uniform(x0+2.0,x1-2.0)
                yt = (y1*(xt-x0) + y0*(x1-xt)) / (x1-x0)
                self.trees.append(self.pointindex(xt,yt))
                if random.random() < APPLE_TREE_RATIO:
                    self.apples.append(self.pointindex(
                        xt+random.uniform(-2.0,2.0),yt+12.0))
        if segheight/seglen < GOODIE_SLOPE_MAX:
            thresh = min(math.atan(GOODIES_PER_FOOT*seglen*f)/f,0.9)
            while random.random() < thresh:
                xt = random.uniform(x0+2.0,x1-2.0)
                yt = (y1*(xt-x0) + y0*(x1-xt)) / (x1-x0)
                self.espressos.append(self.pointindex(xt,yt))
        thresh = min(math.atan(BOUND_NINJAS_PER_FOOT*seglen*f)/f,0.9)
        while random.random() < thresh:
            xt = random.uniform(x0+2.0,x1-2.0)
            yt = (y1*(xt-x0) + y0*(x1-xt)) / (x1-x0)
            self.bound_ninjas.append((self.pointindex(xt,yt),
                                      self.pointindex(x0+0.5,0.0),
                                      self.pointindex(x1-0.5,0.0)))
        thresh = min(math.atan(FREE_NINJAS_PER_FOOT*seglen*f)/f,0.9)
        while random.random() < thresh:
            xt = random.uniform(x0+2.0,x1-2.0)
            yt = (y1*(xt-x0) + y0*(x1-xt)) / (x1-x0)
            self.free_ninjas.append(self.pointindex(xt,yt))


    def add_single_segment(self,x0,y0,x1,y1):
        self.add_multi_segment([x0,x1],[y0,y1])
        

    def add_multi_segment(self,xss,yss):
        logger.debug("        Baseland multisegment: (%g,%g)-(%g,%g)"
                     % (xss[0],yss[0],xss[-1],yss[-1]))
        xss[0],yss[0] = self.adjust_endpoint(xss[0],yss[0])
        xss[-1],yss[-1] = self.adjust_endpoint(xss[-1],yss[-1])

        xground = ET.SubElement(self.xmlroot,"ground")
        xseq = ET.SubElement(xground,"seq")
        si = []
        for x,y in zip(xss,yss):
            si.append(self.pointindex(x,y))
        xseq.text = ",".join(str(s) for s in si)
        xicy = ET.SubElement(xground,"icy")
        xicy.text = "no"
        xjd = ET.SubElement(xground,"jumpdown")
        xjd.text = "no"
        xbanner = ET.SubElement(xground,"banner")
        xbanner.text = "grassy-path.png"

        xlandscape = ET.SubElement(self.xmlroot,"landscape")
        xseq = ET.SubElement(xlandscape,"seq")
        xseq.text = ",".join(str(s) for s in si)
        xscale = ET.SubElement(xlandscape,"scale")
        xscale.text = "0.05"
        xbg = ET.SubElement(xlandscape,"background")
        xbg.text = "rocky-bluff.png"
        
        for x0,y0,x1,y1 in zip(xss[:-1],yss[:-1],xss[1:],yss[1:]):
            self.decorate_segment(x0,y0,x1,y1)


    def set_entrance_point(self,x,y):
        si = self.pointindex(x,y+5)
        ET.SubElement(self.xmlroot,"entrance").text = str(si)
        

    def baseland_gapped(self,x0,y0,x1,y1,entrance=False):
        logger.debug("    Baseland part: (%g,%g)-(%g,%g)" % (x0,y0,x1,y1))
        if random.random() < 0.05:
            self.add_single_segment(x0,y0,x1,y1)
            if entrance:
                xe = x0+MIN_PLATFORM
                ye = (y0*(xe-x0) + y1*(x1-xe)) / (x1-x0)
                self.set_entrance_point(xe,ye)
            return
        avgival = random.uniform(MIN_AVG_SEG_IVAL,MAX_AVG_SEG_IVAL)
        xivals = self.intervalize(x0,x1,avgival,SIGMA_SEG_IVAL,MIN_ABS_SEG_IVAL)
        xss = [x0]+xivals+[x1]
        yivals = [ (y0*(x-x0) + y1*(x1-x)) / (x1-x0) for x in xivals ]
        sigma_hill = random.normalvariate(0.0,SIGMA_SIGMA_HILL)
        yhills = [ y+random.normalvariate(0.0,sigma_hill) for y in yivals ]
        y0h = y0+random.normalvariate(0.0,SIGMA_ENDPOINT)
        y1h = y1+random.normalvariate(0.0,SIGMA_ENDPOINT)
        yss = [y0h]+yhills+[y1h]
        self.add_multi_segment(xss,yss)
        if entrance:
            xe = x0+MIN_PLATFORM
            ye = (y0h*(xe-x0) + yhills[0]*(xivals[0]-xe)) / (xivals[0]-x0)
            self.set_entrance_point(xe,ye)


    def baseland(self,x0,y0,x1,y1,entrance=False):
        logger.debug("New baseland: (%g,%g)-(%g,%g)" % (x0,y0,x1,y1))
        if random.random() < 0.5:
            self.baseland_gapped(x0,y0,x1,y1,entrance)
            return
        ngaps = random.choice(N_GAP_CHOICE)
        gaps = []
        for i in xrange(ngaps):
            while True:
                xc = random.uniform(x0+2*MIN_GAP_WIDTH,x1-2*MIN_GAP_WIDTH)
                wc = random.uniform(MIN_GAP_WIDTH,MAX_GAP_WIDTH)
                xg0c = xc-wc/2
                xg1c = xc+wc/2
                for xg0,xg1 in gaps:
                    if ((xg1c>xg0-MIN_PLATFORM and xg0c<xg0-MIN_PLATFORM)
                        or (xg0c<xg1+MIN_PLATFORM and xg1c>xg1+MIN_PLATFORM)):
                        continue
                gaps.append((xg0c,xg1c))
                break
        gaps.sort()
        xl = x0
        yl = y0
        for i,(xg0,xg1) in enumerate(gaps):
            yg0 = (y0*(xg0-x0) + y1*(x1-xg0)) / (x1-x0)
            yg1 = (y0*(xg1-x0) + y1*(x1-xg1)) / (x1-x0)
            self.baseland_gapped(xl,yl,xg0,yg0,entrance and i==0)
            xl = xg1
            yl = yg1
        self.baseland_gapped(xl,yl,x1,y1,False)

